package	jp.co.sds_corp.report;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Iterator;

import jp.co.sds_corp.util.common.BeanParser;
import jp.co.sds_corp.util.common.CmnLog;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.script.ScriptParserFactory;
import jxl.BooleanCell;
import jxl.Cell;
import jxl.CellFeatures;
import jxl.CellReferenceHelper;
import jxl.CellType;
import jxl.DateCell;
import jxl.LabelCell;
import jxl.NumberCell;
import jxl.Range;
import jxl.Sheet;
import jxl.Workbook;
import jxl.WorkbookSettings;
import jxl.format.BorderLineStyle;
import jxl.format.CellFormat;
import jxl.format.Colour;
import jxl.write.Blank;
import jxl.write.DateTime;
import jxl.write.Formula;
import jxl.write.Label;
import jxl.write.WritableCell;
import jxl.write.WritableCellFeatures;
import jxl.write.WritableCellFormat;
import jxl.write.WritableSheet;
import jxl.write.WritableWorkbook;

/**
<a href="http://www.andykhan.com/jexcelapi/">jexcelapi</a>を使用して、Excel帳票を出力します。
<pre>
&lt;Report
	name="ReportTest"
	type="Jxl"
	template="${Report.dir}/table.xls"
	sheet="Sheet1"&gt;
	&lt;Script ... /&gt;
&lt;/Report&gt;
</pre>
Script要素については、{@link ScriptParserFactory}を参照してください。
*/
public class JxlReport extends AbstractReport
{
	String template;
	String sheetName;
	String orgSheetName;
	String script;
	HashMap<String,WritableSheet> mapTemp = new HashMap<String,WritableSheet>();

	ByteArrayOutputStream	bos;
	protected WritableWorkbook result = null;
	protected WritableSheet sheet = null;
	protected Workbook tmpBook = null;
	protected Sheet orgSheet;
	boolean bPrint = false;
	String charset;

	/**
	* 設定情報で初期化します。
	* @param	config	設定情報
	*/
	public void init( IConfiguration config ) throws Exception {
		super.init( config );
		template = config.getAttributes().get( "template", null );
		if ( getProperty() != null ) {
			template = getProperty().replaceValue( template );
		}
		sheetName = config.getAttributes().get( "sheet", null );
		charset = config.getAttributes().get( "charset", null );

		if ( template != null ) {
			loadFile( template );
			if ( sheetName != null ) {
				loadPage( sheetName );
			}
		}
	}

	/**
	* 処理を終了します。
	*/
	public void exit() throws Exception {
		result = null;
		sheet = null;
		if ( tmpBook != null ) {
			tmpBook.close();
		}
		tmpBook = null;
		orgSheet = null;
	}

	/**
	* 帳票の出力結果をオブジェクトとして取得します。
	* @return	出力結果
	*/
	public Object getResult() {
		return result;
	}

	/**
	* 帳票を出力します。
	*/
	public void print() throws Exception {
		super.print();
		Iterator<String> itr = mapTemp.keySet().iterator();
		while ( itr.hasNext()){
			String name = (String )itr.next();
			int idx = getSheetIndex( result, name );
			if ( isDebug() ) {
				CmnLog.debug( this, "remove template " + name + " (" + idx );
			}
			if ( idx >= 0 ) {
				result.removeSheet( getSheetIndex( result, name ));
			}
		}
		bPrint = true;
	}

	/**
	* 帳票を出力します。
	* @param	out		出力先
	*/
	public void print( OutputStream out ) throws Exception {
		if ( !bPrint ) {
			print();
		}
		result.write();
		result.close();
		byte[] arr = bos.toByteArray();
		out.write( arr );
	}

	/**
	* サーブレットのContentTypeヘッダを取得します。
	* @return	サーブレットのContentTypeヘッダ
	*/
	public String getContentType() {
		return "application/vnd.ms-excel";
	}

	/**
	* ファイルの拡張子を返します。
	* @return   ファイルの拡張子
	*/
	public String getExtension() {
		return "xls";
	}

	private int getSheetIndex( WritableWorkbook book, String name ) {
		int nRet = -1;
		String[] names = book.getSheetNames();
		for ( int i = 0; i < names.length; i++ ) {
			if ( names[i].equals( name )) {
				nRet = i;
				break;
			}
		}
		return nRet;
	}

	/**
	* templateシートの名前を取得します。
	* @param	name	シート名
	*/
	protected String getTemplateName( String name ) {
//		return "__" + name + "__";
		return name;
	}

	/**
	* ファイルを読み込みます。
	* @param	template	ファイル名
	*/
	public void loadFile(String template) throws Exception {
		WorkbookSettings ws = new WorkbookSettings();
//		ws.setIgnoreBlanks( false );
//		ws.setGCDisabled( true );
		if ( charset != null ) {
			ws.setEncoding( charset );
		}
		if ( isDebug()) {
			CmnLog.debug( this, "load file " + template + " " + charset );
		}
		tmpBook = Workbook.getWorkbook( new File( template ), ws );
		bos = new ByteArrayOutputStream();
		result = Workbook.createWorkbook( bos, tmpBook, ws );
		mapTemp.clear();
		for ( int i = 0; i < result.getNumberOfSheets(); i++ ) {
			WritableSheet sheet = result.getSheet( i );
			if ( sheet.getName().equals( "000000" )) {
				if ( isDebug() ) {
					CmnLog.debug( this, "ignore template " + i + ") "
								+ sheet.getName() );
				}
				continue;
			}
			String name = getTemplateName( sheet.getName() );
			if ( isDebug() ) {
				CmnLog.debug( this, "load template " + i + ") "
								+ sheet.getName() + "->" + name );
			}
/*
			sheet.setName( name );
*/
			mapTemp.put( name, sheet );
		}
	}

	/**
	* ページを読み込みます。
	* @param	sheetName	ページ名
	*/
	public void loadPage(String sheetName) throws Exception {
		if ( isDebug() ) {
			CmnLog.debug( this, "load page " + sheetName );
		}
		this.sheetName = sheetName;
		this.orgSheetName = sheetName;
		sheet = result.getSheet( sheetName );
		if ( sheet == null ) {
			throw( new Exception( "Sheet[" + sheetName + "] not found." ));
		}
		orgSheet = tmpBook.getSheet( orgSheetName );
		if ( orgSheet == null ) {
			throw( new Exception( "Sheet[" + orgSheetName + "] not found." ));
		}
		mapTemp.remove( sheetName );
	}
	/**
	* ページに名前をつけて読み込みます。
	* @param	newName	新しいページ名
	* @param	orgName	読み込むページ名
	*/
	public void loadPage(String newName,String orgName) throws Exception {
		orgSheetName = getTemplateName( orgName );
		if ( isDebug() ) {
			CmnLog.debug( this, "load page " + orgSheetName + "->" + newName );
		}
		orgSheet = result.getSheet( orgSheetName );
		if ( orgSheet == null ) {
			throw( new Exception( "Sheet[" + orgSheetName + "] not found." ));
		}
		result.copySheet( orgSheetName, newName, result.getNumberOfSheets() );
		this.sheetName = newName;
		this.sheet = result.getSheet(newName);
	}

	/**
	* 現在処理中のページを、指定されたページに追加します。
	* @param	dstName	追加するページ
	*/
	public void appendTo( String dstName ) throws Exception {
		appendTo( dstName, true );
	}
	/**
	* 現在処理中のページを、指定されたページに追加します。
	* @param	dstName	追加するページ
	* @param	bBreak	改ページフラグ
	*/
	public void appendTo( String dstName, boolean bBreak ) throws Exception {
		if ( sheetName.equals( dstName )) {
			throw( new Exception( "cannot append to same sheet ["
									+ dstName + "]" ));
		}
		WritableSheet dst = result.getSheet( dstName );
		if ( dst == null ) {
			if ( isDebug() ) {
				CmnLog.debug( this, "rename " + sheetName + "->" + dstName );
			}
			int idx = getSheetIndex( result, sheetName );
			sheet.setName( dstName );
			return;
		}
		int nTo = dst.getRows();
		int nRow = sheet.getRows();
		if ( isDebug() ) {
			CmnLog.debug( this, "appendTo " + sheetName + " " + nRow
							+ "->" + dstName + " " + nTo );
		}
		for ( int i = 0; i < nRow; i++ ) {
			Cell[] row = sheet.getRow( i );
			for ( int j = 0; j < row.length; j++ ) {
				WritableCell copy = ((WritableCell)row[j]).copyTo( j, nTo + i );
				copy.setCellFormat( row[j].getCellFormat() );
				dst.addCell( copy );
			}
			dst.setRowView( nTo + i, sheet.getRowView( i ).getSize());
		}
		Range[] range = sheet.getMergedCells();
		for ( int i = 0; i < range.length; i++ ) {
			Cell sta = range[i].getTopLeft();
			Cell end = range[i].getBottomRight();
			dst.mergeCells( sta.getColumn(), sta.getRow() + nTo,
							end.getColumn(), end.getRow() + nTo );
		}
		if ( bBreak ) {
			dst.addRowPageBreak( nTo + nRow );
		}
		int idx = getSheetIndex( result, sheetName );
		result.removeSheet( idx );
	}

	/**
	* 現在処理中のページに改ページを追加します。
	*/
	public void endPage() {
		endPage( sheet.getRows() );
	}
	/**
	* 現在処理中のページに改ページを追加します。
	* @param	row		改ページを追加する行
	*/
	public void endPage( int row ) {
		if ( isDebug()) {
			CmnLog.debug( this, "end page " + row );
		}
		sheet.addRowPageBreak( row );
	}

	/**
	* ページの内容をコピーします。
	* @param	dst		コピー先
	* @param	src		コピー元
	*/
	protected void copySheet( WritableSheet dst, Sheet src ) throws Exception {
		for (int i = 0 ; i < src.getRows(); i++) {
			Cell[] srcRow = src.getRow( i );
			for (int j = 0 ; j < srcRow.length; j++) {
				WritableCell dstCell = dst.getWritableCell( j, i );
				copyCell( dstCell, srcRow[j] );
			}
		}
	}

	/**
	* 行の内容をコピーします。
	* @param	dst		コピー先
	* @param	src		コピー元
	*/
	protected void copyRow( WritableCell[] dst, Cell[] src ) {
		for (int i = 0 ; i < src.length; i++ ) {
			WritableCell dstCell = dst[i];
			Cell srcCell = src[i];
			copyCell( dstCell, srcCell );
		}
	}

	/**
	* セルの内容をコピーします。
	* @param	dst		コピー先
	* @param	src		コピー元
	*/
	protected void copyCell( WritableCell dst, Cell src ) {
		if ( dst == null ) {
			return;
		}
		CellFormat srcFmt = src.getCellFormat();
		if ( srcFmt != null ) {
			CellFormat dstFmt = new WritableCellFormat( srcFmt );
			dst.setCellFormat( dstFmt );
		}
	}

	/**
	* セルに値を設定します。
	* @param	pos		セルの位置
	* @param	data	設定する値
	*/
	public void setCell( String pos, Object data ) throws Exception {
		int col = CellReferenceHelper.getColumn( pos );
		int row = CellReferenceHelper.getRow( pos );
		setCell( col, row, data );
	}

	/**
	* セルに値を設定します。
	* @param	col		セルの桁
	* @param	row		セルの行
	* @param	data	設定する値
	*/
	public void setCell( int col, int row, Object data ) throws Exception {
		WritableCell org = sheet.getWritableCell( col, row );
		CellFormat orgFmt = org.getCellFormat();
		CellType type = org.getType();
		if ( isDebug()) {
			String pos = CellReferenceHelper.getCellReference( col, row );
			String name = ( data == null ? "null" : data.getClass().getName());
			CmnLog.debug( this, "cell " + sheetName + "."
				+ pos + "(" + col + "," + row + ")=" + data
				+ " ;" + name + "->"
				+ type + "(" + org.getClass().getName() + ") " );
		}
		if ( orgFmt == null ) {
			CmnLog.debug( this, "format is null" );
			orgFmt = new WritableCellFormat();
		}
		WritableCell cell = null;
		if ( data instanceof java.lang.String || data == null ) {
			cell = new Label( col, row, (String )data, orgFmt );
		} else
		if ( data instanceof java.util.Date ) {
			cell = new DateTime( col, row, (java.util.Date )data, orgFmt );
		} else
		if ( data instanceof java.lang.Number ) {
			cell = new jxl.write.Number( col, row,
				((java.lang.Number )data).doubleValue(), orgFmt );
		} else
		if ( data instanceof java.lang.Boolean ) {
			cell = new jxl.write.Boolean( col, row,
				((java.lang.Boolean )data).booleanValue(), orgFmt );
		} else {
			throw new IllegalArgumentException("データ型が不正です:"
					+ data.getClass().getName());
		}
		WritableCellFeatures cf = org.getWritableCellFeatures();
		if ( cf != null ) {
			cell.setCellFeatures( cf );
		}
		sheet.addCell( cell );
	}

	/**
	* セルに式を設定します。
	* @param	pos		セルの位置
	* @param	formula	設定する式
	*/
	public void setCellFormula(String pos, String formula) throws Exception {
		int col = CellReferenceHelper.getColumn( pos );
		int row = CellReferenceHelper.getRow( pos );
		setCellFormula(sheet, col, row, formula);
	}

	/**
	* セルに式を設定します。
	* @param	col		セルの桁
	* @param	row		セルの行
	* @param	formula	設定する式
	*/
	public void setCellFormula(int col, int row, String formula) throws Exception {
		setCellFormula( sheet, col, row, formula );
	}

	/**
	* セルに式を設定します。
	* @param	sheet	ページ
	* @param	col		セルの桁
	* @param	row		セルの行
	* @param	formula	設定する式
	*/
	public void setCellFormula(WritableSheet sheet, int col, int row, String formula )
		throws Exception
	{
		WritableCell org = sheet.getWritableCell( col, row );
		CellFormat orgFmt = org.getCellFormat();
		CellType type = org.getType();
		if ( isDebug()) {
			String pos = CellReferenceHelper.getCellReference( col, row );
			CmnLog.debug( this, "formula " + sheetName + "."
				+ pos + "(" + col + "," + row + ")=" + formula );
		}
		if ( orgFmt == null ) {
			CmnLog.debug( this, "format is null" );
			orgFmt = new WritableCellFormat();
		}
		WritableCell cell = null;
		cell = new Formula( col, row, formula, orgFmt );
		sheet.addCell( cell );
	}

	/**
	* セルに罫線を設定します。
	* @param	pos		セルの位置
	* @param	location	罫線の位置
	* @param	border	設定する罫線
	*/
	public void setBorder( String pos, int location, Border border )
		throws Exception
	{
		int col = CellReferenceHelper.getColumn( pos );
		int row = CellReferenceHelper.getRow( pos );
		setBorder( col, row, location, border );
	}
	/**
	* セルに式を設定します。
	* @param	col		セルの桁
	* @param	row		セルの行
	* @param	location	罫線の位置
	* @param	border	設定する罫線
	*/
	public void setBorder( int col, int row, int location, Border border )
		throws Exception
	{
		if ( isDebug()) {
			String pos = CellReferenceHelper.getCellReference( col, row );
			CmnLog.debug( this, "border " + sheetName + "."
				+ pos + "(" + col + "," + row + ")." + location + "=" + border);
		}
		jxl.format.Border loc = null;
		switch ( location ) {
		case Border.TOP: loc = jxl.format.Border.TOP; break;
		case Border.BOTTOM: loc = jxl.format.Border.BOTTOM; break;
		case Border.LEFT: loc = jxl.format.Border.LEFT; break;
		case Border.RIGHT: loc = jxl.format.Border.RIGHT; break;
		}
		BorderLineStyle style = convertBorder( border );
		WritableCell cell = sheet.getWritableCell( col, row );
		if ( cell.getType().equals( CellType.EMPTY )) {
			cell = new Label( col, row, "" );
			sheet.addCell( cell );
		}
		WritableCellFormat orgFmt = (WritableCellFormat )cell.getCellFormat();
		WritableCellFormat newFmt = new WritableCellFormat();
		BeanParser parser = new BeanParser( orgFmt );
		parser.copyTo( newFmt );
		copyBorder( newFmt, orgFmt );
		newFmt.setBorder( loc, style );
		cell.setCellFormat( newFmt );
	}

	protected BorderLineStyle convertBorder( Border border ) {
		BorderLineStyle ret;
		switch ( border.getType() ){
		case Border.TYPE_LINE:
			ret = BorderLineStyle.THIN;
			break;
		case Border.TYPE_DOT:
			ret = BorderLineStyle.DOTTED;
			break;
		case Border.TYPE_DASH_DOT:
			ret = BorderLineStyle.DASH_DOT;
			break;
		case Border.TYPE_DOUBLE:
			ret = BorderLineStyle.DOUBLE;
			break;
		case Border.TYPE_NONE:
			default:
			ret = BorderLineStyle.NONE;
		}
		return ret;
	}

	protected void copyBorder( WritableCellFormat dst, CellFormat src )
		throws Exception
	{
		Colour col;
		col = src.getBorderColour( jxl.format.Border.TOP );
		dst.setBorder( jxl.format.Border.TOP,
					src.getBorder( jxl.format.Border.TOP ), col );
		col = src.getBorderColour( jxl.format.Border.BOTTOM );
		dst.setBorder( jxl.format.Border.BOTTOM,
					src.getBorder( jxl.format.Border.BOTTOM ), col );
		col = src.getBorderColour( jxl.format.Border.LEFT );
		dst.setBorder( jxl.format.Border.LEFT,
					src.getBorder( jxl.format.Border.LEFT ), col );
		col = src.getBorderColour( jxl.format.Border.RIGHT );
		dst.setBorder( jxl.format.Border.RIGHT,
					src.getBorder( jxl.format.Border.RIGHT ), col );
	}

	/**
	* セルの値をコピーします。
	* @param	fpos	コピー元のセルの位置
	* @param	tpos	コピー先のセルの位置
	*/
	public void copyCell( String fpos, String tpos ) throws Exception {
		int fcol = CellReferenceHelper.getColumn( fpos );
		int frow = CellReferenceHelper.getRow( fpos );
		int tcol = CellReferenceHelper.getColumn( tpos );
		int trow = CellReferenceHelper.getRow( tpos );
		copyCell( fcol, frow, tcol, trow );
	}
	/**
	* セルの値をコピーします。
	* @param	fcol	コピー元のセルの桁
	* @param	frow	コピー元のセルの行
	* @param	tcol	コピー先のセルの桁
	* @param	trow	コピー先のセルの行
	*/
	public void copyCell( int fcol, int frow, int tcol, int trow )
		throws Exception
	{
		if ( isDebug()) {
			String fpos = CellReferenceHelper.getCellReference( fcol, frow );
			String tpos = CellReferenceHelper.getCellReference( tcol, trow );
			CmnLog.debug( this, "copy "
				+ orgSheetName + "." + fpos + "(" + fcol + "," + frow + ")->"
				+ sheetName + "." + tpos + "(" + tcol + "," + trow + ")" );
		}
		Cell srcCell = orgSheet.getCell( fcol, frow );
		CellFormat srcFmt = srcCell.getCellFormat();
		CellFormat dstFmt = null;
		if ( srcFmt != null ) {
			dstFmt = new WritableCellFormat( srcFmt );
		}
		WritableCell dstCell = null;
		if ( srcCell.getType().equals( CellType.LABEL )) {
			dstCell = new Label( tcol, trow,
				((LabelCell )srcCell).getString(), srcFmt );
		} else
		if ( srcCell.getType().equals( CellType.DATE )) {
			dstCell = new DateTime( tcol, trow,
				((DateCell )srcCell).getDate(), srcFmt );
		} else
		if ( srcCell.getType().equals( CellType.NUMBER )) {
			dstCell = new jxl.write.Number( tcol, trow,
				((NumberCell )srcCell).getValue(), srcFmt );
		} else
		if ( srcCell.getType().equals( CellType.BOOLEAN )) {
			dstCell = new jxl.write.Boolean( tcol, trow,
				((BooleanCell )srcCell).getValue(), srcFmt );
		} else {
			dstCell = new Blank( tcol, trow, srcFmt );
		}
		CellFeatures cf = srcCell.getCellFeatures();
		if ( cf != null ) {
			dstCell.setCellFeatures( new WritableCellFeatures( cf ) );
		}
		sheet.addCell( dstCell );
	}

	/**
	* セルをクリアします。
	* @param    pos		セルの位置
	*/
	public void clearCell( String pos ) throws Exception {
		int col = CellReferenceHelper.getColumn( pos );
		int row = CellReferenceHelper.getRow( pos );
		clearCell( col, row );
	}
	/**
	* セルをクリアします。
	* @param	col		コピー元のセルの桁
	* @param	row		コピー元のセルの行
	*/
	public void clearCell( int col, int row ) throws Exception {
		if ( isDebug()) {
			String pos = CellReferenceHelper.getCellReference( col, row );
			CmnLog.debug( this, "clear " + sheetName + "."
				+ pos + "(" + col + "," + row + ")" );
		}
		setCell( col, row, "" );
		WritableCell dstCell = sheet.getWritableCell( col, row );
		dstCell.setCellFormat( new WritableCellFormat() );
	}

	/**
	* 現在のページに行を追加します。
	* @param	row		追加する位置
	*/
	public void addRow( int row ) throws Exception {
		if ( isDebug()) {
			CmnLog.debug( this, "add row " + row );
		}
		sheet.insertRow( row );
	}
	/**
	* 現在のページに、templateから書式をコピーして行を追加します。
	* @param	dst		追加する位置
	* @param	src		書式の行の位置
	*/
	public void addRow( int dst, int src ) throws Exception {
		Cell[] srcRow = orgSheet.getRow( src );
		addRow( dst, src, 0, srcRow.length );
	}
	/**
	* 現在のページに、templateから書式をコピーして行を追加します。
	* @param	dst		追加する位置
	* @param	src		書式の行の位置
	* @param	colFr	書式の桁開始位置
	* @param	colTo	書式の桁終了位置
	*/
	public void addRow( int dst, int src, int colFr, int colTo )
		throws Exception
	{
		if ( isDebug()) {
			CmnLog.debug( this, "copy row " + src + "->"
								+ dst + "[" + colFr + "," + colTo + "]" );
		}
		Cell[] srcRow = orgSheet.getRow( src );
		for ( int i = colFr; i < colTo; i++ ) {
			WritableCell srcCell = sheet.getWritableCell( i, src );
			WritableCell dstCell = srcCell.copyTo( i, dst );
			sheet.addCell( dstCell );
		}
		Range[] range = sheet.getMergedCells();
		for ( int i = 0; i < range.length; i++ ) {
			Cell sta = range[i].getTopLeft();
			Cell end = range[i].getBottomRight();
			if ( sta.getRow() == src ) {
				sheet.mergeCells( sta.getColumn(), dst,
								end.getColumn(), dst + end.getRow() - sta.getRow() );
			}
		}
	}
	/**
	* 行の処理を終了します。
	*/
	public void endRow() throws Exception {
	}

}
